/*
	$Id: BufferedInputStream.c 117 2006-03-07 11:53:10Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <stdlib.h>
#include <string.h>
#include "tricks.h"
#include "BufferedInputStream.h"
#include "impl/Object.h"
#include "impl/InputStream.h"

/* __final__ */ struct impl_BufferedInputStream {
    union {
	struct impl_Object object;
	struct impl_InputStream inputStream;
    } super;
    struct impl_InputStream *source;
    size_t size;
    unsigned char *data;
    size_t bufferSize;
    unsigned char buffer[0];
};

union BIS {
    struct impl_InputStream *inputStream;
    struct impl_BufferedInputStream *impl;
} TRANSPARENT_UNION;

static ssize_t
BIS_read(union BIS this, void *dataTop, size_t size)
{
    struct impl_BufferedInputStream *impl = this.impl;
    struct impl_InputStream *source = impl->source;
    unsigned char *data = dataTop;
    ssize_t s, v;

    if (impl->size > size) {
	memcpy(data, impl->data, size);
	impl->data += size;
	impl->size -= size;
	return ((data + size) - (unsigned char *)dataTop);
    }
    if (impl->size > 0) {
	memcpy(data, impl->data, impl->size);
	data += impl->size;
	size -= impl->size;
	impl->data = impl->buffer + impl->bufferSize;
	impl->size = 0;
    }
    v = data - (unsigned char *)dataTop;
    if (size == 0)
	return v;
    if (size > impl->bufferSize)
	return ((s = source->read(source, data, size)) < 0) ? s : v + s;
    if ((s = source->read(source, impl->buffer, impl->bufferSize)) < 0)
	return s;
    if (s == 0)
	return v;
    impl->data = impl->buffer;
    impl->size = s;
    if (size > (size_t)s)
	size = s;
    memcpy(data, impl->data, size);
    impl->data += size;
    impl->size -= size;
    return v + size;
}

static int
BIS_close(union BIS this)
{
    struct impl_BufferedInputStream *impl = this.impl;
    struct impl_InputStream *source = impl->source;
    int s;

    s = source->close(source);
    free(impl);
    return s;
}

union BufferedInputStream
BufferedInputStream_create(union InputStream in)
{
    struct impl_InputStream *source = in.impl;
    struct impl_BufferedInputStream *impl;
    static const size_t size = 4096;

    if (source == NULL) {
	return (union BufferedInputStream)null;
    }
    if ((impl = malloc(sizeof(*impl) + size)) == NULL) {
	source->close(source);
	return (union BufferedInputStream)null;
    }
    impl->super.inputStream.read = BIS_read;
    impl->super.inputStream.close = BIS_close;
    impl->source = source;
    impl->size = 0;
    impl->data = impl->buffer;
    impl->bufferSize = size;
    return (union BufferedInputStream)impl;
}
